--
--	WhenLevel
--	Written by scott@mithrandir.com
--
--	Version History
--		001	Initial version that records when leveling is done.
--		002	Added command line to output leveling information.
--		003	Converted to using strings in preparation for localization.
--		004	Computes maximum level instead of hardcoding 80.
--		005	Converted to using a cross-character data store and storing location.
--			Made it so the output appears in a scroll view in the standard Blizzard
--			configuration panel instead of the default chat window.
--		006	Widened the configuration to give more space to the zone.
--			Updated to run on release 4.0.1.
--
--	UTF-8 file
--

WhenLevel_File_Version = 006

if nil == WhenLevel or WhenLevel.versionNumber < WhenLevel_File_Version then

	--	This is the data that is written to disk that records each level date and time played.
	WhenLevelData = { }

	--	Starting in version 005 the data is stored in one central location so character leveling can be compared.
	WhenLevelDatabase = { }

	--	The actual addon itself.
	--	Conceptually a very simple addon that creates a tooltip that is used to receive events for the addon.
	--	The tooltip is hidden and is not used for any other purpose.  A slash command is registered so the
	--	user can request the report of leveling information.  Two events are of importance -- PLAYER_LEVEL_UP
	--	and TIME_PLAYED_MSG.  When the addon receives notification that PLAYER_LEVEL_UP it ensures the server
	--	is asked for how much time is played, causing TIME_PLAYED_MSG to eventually be sent.  Note that within
	--	the processing of the PLAYER_LEVEL_UP event the call to UnitLevel() will not return the new level, so
	--	we use the first argument in the PLAYER_LEVEL_UP event to record this information.  Also note that upon
	--	entry into the world the server now reports the time played so we also receive TIME_PLAYED_MSG.  In
	--	processing TIME_PLAYED_MSG we record the server time and the total time played if we have not already
	--	recorded this information for the current character level.  Since we only record the time played at a
	--	level once, it is safe for the user to use the /played slash command without harming the WhenLevel data.
	--	Since a lot of characters will first use WhenLevel after level one, it is a special case and noted in
	--	the /whenlevel output data.
	WhenLevel = {

		versionNumber = WhenLevel_File_Version,
		began = nil,
		currentlyProcessingLevel,
		eventDispatch = {			-- table of functions whose keys are the events
			['PLAYER_ENTERING_WORLD'] = function(self, frame)
				self.locale = GetLocale()
				-- if the locale is not found to be supported default to enUS
				if nil == self.s[self.locale] then
					self.locale = 'enUS'
				end
				self.currentlyProcessingLevel = UnitLevel("player")
				frame:RegisterEvent("PLAYER_LEVEL_UP")
				frame:RegisterEvent("TIME_PLAYED_MSG")
			end,
			['PLAYER_LEVEL_UP'] = function(self, frame, arg1)
				--	arg1 tells us the new level, but we just cause another event to be posted
				self.currentlyProcessingLevel = arg1
				RequestTimePlayed()
			end,
			['TIME_PLAYED_MSG'] = function(self, frame, arg1, arg2)
				--	arg1 tells us total time
				--	arg2 tells us total time at level
				local currentLevel = self.currentlyProcessingLevel
				self.began = true
				if (nil == WhenLevelDatabase[self.playerRealm]) then
					WhenLevelDatabase[self.playerRealm] = { }
				end
				if (nil == WhenLevelDatabase[self.playerRealm][self.playerName]) then
					WhenLevelDatabase[self.playerRealm][self.playerName] = { }
					WhenLevelDatabase[self.playerRealm][self.playerName]["class"] = self.playerClass
					local whereToStop = UnitLevel('player')
					for i = 1, whereToStop, 1 do
						if (nil ~= WhenLevelData[i]) then
							WhenLevelDatabase[self.playerRealm][self.playerName][i] = { }
							WhenLevelDatabase[self.playerRealm][self.playerName][i]["totalPlayed"] = WhenLevelData[i]["totalPlayed"]
							WhenLevelDatabase[self.playerRealm][self.playerName][i]["serverTime"] = WhenLevelData[i]["serverTime"]
						end
					end
				end
				if (nil == WhenLevelDatabase[self.playerRealm][self.playerName][currentLevel]) then
					local hour, minute = GetGameTime()
					local weekday, month, day, year = CalendarGetDate()
					local x, y = GetPlayerMapPosition("player")
					WhenLevelDatabase[self.playerRealm][self.playerName][currentLevel] = { }
					WhenLevelDatabase[self.playerRealm][self.playerName][currentLevel]["totalPlayed"] = arg1
					WhenLevelDatabase[self.playerRealm][self.playerName][currentLevel]["serverTime"] = string.format("%4d-%02d-%02d %02d:%02d", year, month, day, hour, minute)
					WhenLevelDatabase[self.playerRealm][self.playerName][currentLevel]["zone"] = GetRealZoneText()
					WhenLevelDatabase[self.playerRealm][self.playerName][currentLevel]["subzone"] = GetSubZoneText()
					WhenLevelDatabase[self.playerRealm][self.playerName][currentLevel]["x"] = x
					WhenLevelDatabase[self.playerRealm][self.playerName][currentLevel]["y"] = y
				end
			end,
			['VARIABLES_LOADED'] = function(self, frame)
				--	Ensure the tooltip is not messed up
				if not com_mithrandir_whenLevelTooltip:IsOwned(UIParent) then
					com_mithrandir_whenLevelTooltip:SetOwner(UIParent, "ANCHOR_NONE")
				end
				frame:RegisterEvent("PLAYER_ENTERING_WORLD")
			end,
			},
		locale,
		playerClass,
		playerName,
		playerRealm,
		rowCount = 25,
		--	This section has the localization information for all the output text that the addon generates.
		--	Note that it is done this way solely to minimize the namespace pollution this addon causes.
		--	The downfall of this method is a larger runtime size since each of the localizations is kept
		--	in memory, and the need for detecting whether the locale the user is using is actually supported.
		--	We could eliminate the runtime size issue by eliminating the non-used locales, but that would
		--	require a little more code and the desire is smaller.  The number of strings used in this addon
		--	do not require that much memory, so the desire for less namespace pollution wins.
		s = {
			['enUS'] = {
				FIRST_USED          = 'First Used',
				DAYS                = 'd',
				HOURS               = 'h',
				MINUTES             = 'm',
				SECONDS             = 's',
				LEVEL               = 'Level',
				DATE                = 'Date',
                                ZONE                = 'Zone',
                                TOTAL_PLAYED        = 'Total Played',
				LEVEL_PLAYED        = 'Level Played',
				},
			['deDE'] = {
				FIRST_USED          = 'First Used',
				DAYS                = 'T',
				HOURS               = 'H',
				MINUTES             = 'M',
				SECONDS             = 'S',
				LEVEL               = 'Level',
				DATE                = 'Date',
                                ZONE                = 'Zone',
                                TOTAL_PLAYED        = 'Total Played',
				LEVEL_PLAYED        = 'Level Played',
				},
			['enGB'] = {
				FIRST_USED          = 'First Used',
				DAYS                = 'd',
				HOURS               = 'h',
				MINUTES             = 'm',
				SECONDS             = 's',
				LEVEL               = 'Level',
				DATE                = 'Date',
                                ZONE                = 'Zone',
                                TOTAL_PLAYED        = 'Total Played',
				LEVEL_PLAYED        = 'Level Played',
				},
			['esES'] = {
				FIRST_USED          = 'First Used',
				DAYS                = 'd',
				HOURS               = 'h',
				MINUTES             = 'm',
				SECONDS             = 's',
				LEVEL               = 'Level',
				DATE                = 'Date',
                                ZONE                = 'Zone',
                                TOTAL_PLAYED        = 'Total Played',
				LEVEL_PLAYED        = 'Level Played',
				},
			['esMX'] = {
				FIRST_USED          = 'First Used',
				DAYS                = 'd',
				HOURS               = 'h',
				MINUTES             = 'm',
				SECONDS             = 's',
				LEVEL               = 'Level',
				DATE                = 'Date',
                                ZONE                = 'Zone',
                                TOTAL_PLAYED        = 'Total Played',
				LEVEL_PLAYED        = 'Level Played',
				},
			['frFR'] = {
				FIRST_USED          = 'First Used',
				DAYS                = 'j',
				HOURS               = 'h',
				MINUTES             = 'm',
				SECONDS             = 's',
				LEVEL               = 'Level',
				DATE                = 'Date',
                                ZONE                = 'Zone',
                                TOTAL_PLAYED        = 'Total Played',
				LEVEL_PLAYED        = 'Level Played',
				},
			['koKR'] = {
				FIRST_USED          = 'First Used',
				DAYS                = 'd',
				HOURS               = 'h',
				MINUTES             = 'm',
				SECONDS             = 's',
				LEVEL               = 'Level',
				DATE                = 'Date',
                                ZONE                = 'Zone',
                                TOTAL_PLAYED        = 'Total Played',
				LEVEL_PLAYED        = 'Level Played',
				},
			['ruRU'] = {
				FIRST_USED          = 'First Used',
				DAYS                = 'd',
				HOURS               = 'h',
				MINUTES             = 'm',
				SECONDS             = 's',
				LEVEL               = 'Level',
				DATE                = 'Date',
                                ZONE                = 'Zone',
                                TOTAL_PLAYED        = 'Total Played',
				LEVEL_PLAYED        = 'Level Played',
				},
			['zhCN'] = {
				FIRST_USED          = 'First Used',
				DAYS                = 'd',
				HOURS               = 'h',
				MINUTES             = 'm',
				SECONDS             = 's',
				LEVEL               = 'Level',
				DATE                = 'Date',
                                ZONE                = 'Zone',
                                TOTAL_PLAYED        = 'Total Played',
				LEVEL_PLAYED        = 'Level Played',
				},
			['zhTW'] = {
				FIRST_USED          = 'First Used',
				DAYS                = 'd',
				HOURS               = 'h',
				MINUTES             = 'm',
				SECONDS             = 's',
				LEVEL               = 'Level',
				DATE                = 'Date',
                                ZONE                = 'Zone',
                                TOTAL_PLAYED        = 'Total Played',
				LEVEL_PLAYED        = 'Level Played',
				},
			},

		FormattedSeconds = function(self, totalSeconds)
			local days = math.floor(totalSeconds / 86400)
			local remainder = totalSeconds - (days * 86400)
			local hours = math.floor(remainder / 3600)
			remainder = remainder - (hours * 3600)
			local minutes = math.floor(remainder / 60)
			local seconds = remainder - (minutes * 60)
			local retval = ""

			if (days > 0) then retval = retval .. days .. self.s[self.locale]['DAYS'] end
			if (hours > 0 or days > 0) then retval = retval .. hours .. self.s[self.locale]['HOURS'] end
			if (minutes > 0 or hours > 0 or days > 0) then retval = retval .. minutes .. self.s[self.locale]['MINUTES'] end
			retval = retval .. seconds .. self.s[self.locale]['SECONDS']
			return retval
		end,

		ScrollBar_Update = function(self, scroller)
			local line
			local linePlusOffset
			local whereToStop = UnitLevel('player')
			local date, zone, totalPlayed, levelPlayed
			local firstLevelEncounterd = nil
			local totalRowCount = 50

			-- We need to determine the range that the scroll bar needs to cover because we are not guaranteed that
			-- the start will be level 1 since toons may have not used this since their creation.
			for i = 1, whereToStop, 1 do
				if (nil ~= WhenLevelDatabase[self.playerRealm][self.playerName][i]) then
					if (nil == firstLevelEncountered) then
						firstLevelEncountered = i
					end
				end
			end

			totalRowCount = 1 + whereToStop - firstLevelEncountered
			FauxScrollFrame_Update(_G["com_mithrandir_whenLevelScrollBar"], totalRowCount, self.rowCount, 16)

			for line = 1, self.rowCount do
				linePlusOffset = firstLevelEncountered - 1 + line + FauxScrollFrame_GetOffset(_G["com_mithrandir_whenLevelScrollBar"])
				if linePlusOffset <= whereToStop then
					_G["com_mithrandir_whenLevelButton"..line.."Level"]:SetText(linePlusOffset)
					date = ""
					zone = ""
					totalPlayed = ""
					levelPlayed = ""					
					if (nil ~= WhenLevelDatabase[self.playerRealm][self.playerName][linePlusOffset]) then
						if (linePlusOffset ~= firstLevelEncountered or firstLevelEncountered == 1) then
							date = WhenLevelDatabase[self.playerRealm][self.playerName][linePlusOffset]["serverTime"]
							if (nil ~= WhenLevelDatabase[self.playerRealm][self.playerName][linePlusOffset]["zone"]) then
								zone = WhenLevelDatabase[self.playerRealm][self.playerName][linePlusOffset]["zone"]
								if (nil ~= WhenLevelDatabase[self.playerRealm][self.playerName][linePlusOffset]["subzone"]) then
									if ("" ~= WhenLevelDatabase[self.playerRealm][self.playerName][linePlusOffset]["subzone"]) then
										zone = zone .. " (" .. WhenLevelDatabase[self.playerRealm][self.playerName][linePlusOffset]["subzone"] .. ")"
									end
								end
							end
							totalPlayed = self:FormattedSeconds(WhenLevelDatabase[self.playerRealm][self.playerName][linePlusOffset]["totalPlayed"])
							if (linePlusOffset < whereToStop) then
								local totalSeconds = WhenLevelDatabase[self.playerRealm][self.playerName][linePlusOffset + 1]["totalPlayed"] - WhenLevelDatabase[self.playerRealm][self.playerName][linePlusOffset]["totalPlayed"]
								levelPlayed = string.format(self:FormattedSeconds(totalSeconds))
							end
						else
							date = string.format(self.s[self.locale]['FIRST_USED'])
						end
					end
					_G["com_mithrandir_whenLevelButton"..line.."Date"]:SetText(date)
					_G["com_mithrandir_whenLevelButton"..line.."Zone"]:SetText(zone)
					_G["com_mithrandir_whenLevelButton"..line.."TotalPlayed"]:SetText(totalPlayed)
					_G["com_mithrandir_whenLevelButton"..line.."LevelPlayed"]:SetText(levelPlayed)
					_G["com_mithrandir_whenLevelButton"..line]:Show()
				else
					_G["com_mithrandir_whenLevelButton"..line]:Hide()
				end
			end
		end,

		SlashCommand = function(self, msg)
			InterfaceOptionsFrame_OpenToCategory("WhenLevel")
		end,

		Tooltip_OnEvent = function(self, frame, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9)
			if self.eventDispatch[event] then
				self.eventDispatch[event](self, frame, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9)
			end
		end,

		ConfigFrame_OnLoad = function(self, panel)
			panel.name = "WhenLevel"
			panel:Hide()
			InterfaceOptions_AddCategory(panel)
		end,

		ConfigFrame_OnShow = function(self, panel)
			--	Create the user interface
			local frame = CreateFrame("Frame", "com_mithrandir_whenLevelFrame", panel)
			frame:SetHeight(432)
			frame:SetWidth(660)
			frame:SetPoint("TOPLEFT", panel, "TOPLEFT")

			local columnHeader = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
			columnHeader:SetPoint("TOPLEFT", frame, "TOPLEFT", 7, -8)
			columnHeader:SetText(self.s[self.locale]['LEVEL'])

			columnHeader = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
			columnHeader:SetPoint("TOPLEFT", frame, "TOPLEFT", 50, -8)
			columnHeader:SetText(self.s[self.locale]['DATE'])

			columnHeader = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
			columnHeader:SetPoint("TOPLEFT", frame, "TOPLEFT", 183, -8)
			columnHeader:SetText(self.s[self.locale]['ZONE'])

			columnHeader = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
			columnHeader:SetPoint("TOPLEFT", frame, "TOPLEFT", 434, -8)
			columnHeader:SetText(self.s[self.locale]['TOTAL_PLAYED'])

			columnHeader = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
			columnHeader:SetPoint("TOPLEFT", frame, "TOPLEFT", 545, -8)
			columnHeader:SetText(self.s[self.locale]['LEVEL_PLAYED'])

			local scrollFrame = CreateFrame("ScrollFrame", "com_mithrandir_whenLevelScrollBar", frame, "FauxScrollFrameTemplate")
			scrollFrame:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, -18)
			scrollFrame:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -30, 8)
			--	Note the use of the following getglobal to append a "ScrollBar" to the name of the UI element we created.  This
			--	is done because we need access to an internal UI element to get its value, otherwise scrolling will not happen.
			--	If the ScrollFrame were created in XML we would be fine, but in LUA we are in trouble and need to hack.
			scrollFrame:SetScript("OnVerticalScroll", function() FauxScrollFrame_OnVerticalScroll(scrollFrame, _G["com_mithrandir_whenLevelScrollBarScrollBar"]:GetValue(), 16, WhenLevelScrollBar_Update); end)

			local newButton
			local button = CreateFrame("Button", "com_mithrandir_whenLevelButton1", frame, "com_mithrandir_whenLevelButtonTemplate")
			button:SetPoint("TOPLEFT", scrollFrame, "TOPLEFT", 8, 0)

			for i = 2, self.rowCount do
				newButton = CreateFrame("Button", "com_mithrandir_whenLevelButton"..i, frame, "com_mithrandir_whenLevelButtonTemplate")
				newButton:SetPoint("TOPLEFT", button, "BOTTOMLEFT")
				button = newButton
			end

			WhenLevelScrollBar_Update(scrollFrame)
		end,


		Tooltip_OnLoad = function(self, frame)
			self.playerRealm = GetRealmName()
			self.playerName = UnitName('player')
			_, self.playerClass = UnitClass('player')
			SlashCmdList["WHENLEVEL"] = function(msg)
				self:SlashCommand(msg)
			end
			SLASH_WHENLEVEL1 = "/whenlevel"
			frame:SetScript("OnEvent", function(frame, event, ...) self:Tooltip_OnEvent(frame, event, ...) end)
			frame:RegisterEvent("VARIABLES_LOADED")
		end,

		}

end

--	This function exists because it appears calling the actual function within WhenLevel from FauxScrollFrame_OnVerticalScroll()
--	causes problems such that the low and high end values in the scroll range change in unpredictable ways.  Therefore, we just
--	pass on to the actual function here and all seems to work well.
function WhenLevelScrollBar_Update(scroller)
	WhenLevel:ScrollBar_Update(scroller)
end
